import json
import re
from typing import Any, Dict, Type, Union

from wicker.core.errors import WickerSchemaException
from wicker.schema import codecs, schema
from wicker.schema.schema import PRIMARY_KEYS_TAG

JSON_SCHEMA_VERSION = 2
DATASET_ID_REGEX = re.compile(r"^(?P<name>[0-9a-zA-Z_]+)/v(?P<version>[0-9]+\.[0-9]+\.[0-9]+)$")


def dumps(schema: schema.DatasetSchema, pretty: bool = True) -> str:
    """Dumps a schema as JSON

    :param schema: schema to dump
    :type schema: schema.DatasetSchema
    :param pretty: whether to dump the schema as a prettified JSON, defaults to False
    :type pretty: bool, optional
    :return: JSON string
    :rtype: str
    """
    visitor = AvroDatasetSchemaSerializer()
    jdata = schema.schema_record._accept_visitor(visitor)
    jdata.update(jdata["type"])
    jdata["_json_version"] = JSON_SCHEMA_VERSION
    if pretty:
        return json.dumps(jdata, sort_keys=True, indent=4)
    return json.dumps(jdata)


def loads(schema_str: str, treat_objects_as_bytes: bool = False) -> schema.DatasetSchema:
    """Loads a DatasetSchema from a JSON str

    :param schema_str: JSON string
    :param treat_objects_as_bytes: If set, don't load codecs for object types, instead just treat them as bytes.
      This is useful for code that needs to work with datasets in a generic way, but does not need to actually decode
      the data.
    :return: Deserialized DatasetSchema object
    """
    # Parse string as JSON
    try:
        schema_dict = json.loads(schema_str)
    except json.decoder.JSONDecodeError:
        raise WickerSchemaException(f"Unable load string as JSON: {schema_str}")

    # Construct the DatasetSchema
    try:
        fields = [_loads(d, treat_objects_as_bytes=treat_objects_as_bytes) for d in schema_dict["fields"]]
        return schema.DatasetSchema(
            fields=fields,
            primary_keys=json.loads(schema_dict.get(PRIMARY_KEYS_TAG, "[]")),
            allow_empty_primary_keys=True,  # For backward compatibility. Clean me AVSW-78939.
        )
    except KeyError:
        raise WickerSchemaException(f"Malformed serialization of DatasetSchema: {schema_str}")


def _loads(schema_dict: Dict[str, Any], treat_objects_as_bytes: bool) -> schema.SchemaField:
    """Recursively parses a schema dictionary into the appropriate SchemaField"""
    type_ = schema_dict["type"]
    required = True

    # Nullable types are represented in Avro schemas as Union (list) types with "null" as the
    # first element (by convention)
    if isinstance(type_, list):
        assert type_[0] == "null"
        required = False
        type_ = type_[1]

    if isinstance(type_, dict) and type_["type"] == "record":
        return schema.RecordField(
            name=schema_dict["name"],
            fields=[_loads(f, treat_objects_as_bytes=treat_objects_as_bytes) for f in type_["fields"]],
            description=schema_dict["_description"],
            required=required,
        )
    if isinstance(type_, dict) and type_["type"] == "array":
        # ArrayField columns are limited to contain only dicts or simple types (and not nested arrays)
        element_dict = type_.copy()
        element_dict["type"] = type_["items"]
        element_field = _loads(element_dict, treat_objects_as_bytes=treat_objects_as_bytes)
        return schema.ArrayField(
            element_field=element_field,
            required=required,
        )
    return _loads_base_types(type_, required, schema_dict, treat_objects_as_bytes=treat_objects_as_bytes)


def _loads_base_types(
    type_: str, required: bool, schema_dict: Dict[str, Any], treat_objects_as_bytes: bool = False
) -> schema.SchemaField:
    if type_ == "int":
        return schema.IntField(
            name=schema_dict["name"],
            description=schema_dict["_description"],
            required=required,
        )
    elif type_ == "long":
        return schema.LongField(
            name=schema_dict["name"],
            description=schema_dict["_description"],
            required=required,
        )
    elif type_ == "string":
        return schema.StringField(
            name=schema_dict["name"],
            description=schema_dict["_description"],
            required=required,
        )
    elif type_ == "boolean":
        return schema.BoolField(
            name=schema_dict["name"],
            description=schema_dict["_description"],
            required=required,
        )
    elif type_ == "float":
        return schema.FloatField(
            name=schema_dict["name"],
            description=schema_dict["_description"],
            required=required,
        )
    elif type_ == "double":
        return schema.DoubleField(
            name=schema_dict["name"],
            description=schema_dict["_description"],
            required=required,
        )
    elif type_ == "bytes":
        l5ml_metatype = schema_dict["_l5ml_metatype"]
        if l5ml_metatype == "object":
            codec_name = schema_dict["_codec_name"]
            if treat_objects_as_bytes:
                codec: codecs.Codec = _PassThroughObjectCodec(codec_name, json.loads(schema_dict["_codec_params"]))
            else:
                try:
                    codec_cls: Type[codecs.Codec] = codecs.Codec.codec_registry[codec_name]
                except KeyError:
                    raise WickerSchemaException(
                        f"Could not find a registered codec with name {codec_name} "
                        f"for field {schema_dict['name']}. Please define a subclass of ObjectField.Codec and define "
                        "the codec_name static method."
                    )
                codec = codec_cls.load_codec_from_dict(json.loads(schema_dict["_codec_params"]))
            return schema.ObjectField(
                name=schema_dict["name"],
                codec=codec,
                description=schema_dict["_description"],
                required=required,
                is_heavy_pointer=schema_dict.get("_is_heavy_pointer", False),
            )
        raise WickerSchemaException(f"Unhandled _l5ml_metatype for avro bytes type: {l5ml_metatype}")
    raise WickerSchemaException(f"Unhandled type: {type_}")


class _PassThroughObjectCodec(codecs.Codec):
    """The _PassThroughObjectCodec class is a placeholder for any object codec, it is used when we need to parse
    any possible schema properly, but we don't need to read the data. This codec does not decode/encode the data,
    instead is just acts as an identity function. However, it keeps track of all the attributes of the original codec
    as they were stored in the loaded schema, so that if we save that schema again, we do not lose any information.
    """

    def __init__(self, codec_name: str, codec_attributes: Dict[str, Any]):
        self._codec_name_value = codec_name
        self._codec_attributes = codec_attributes

    @staticmethod
    def _codec_name() -> str:
        # The static method does not make any sense here, because this class is a placeholder for any codec class.
        # Instead we will overload the get_codec_name function.
        return ""

    def get_codec_name(self) -> str:
        return self._codec_name_value

    def save_codec_to_dict(self) -> Dict[str, Any]:
        return self._codec_attributes

    @staticmethod
    def load_codec_from_dict(data: Dict[str, Any]) -> codecs.Codec:
        pass

    def validate_and_encode_object(self, obj: bytes) -> bytes:
        return obj

    def decode_object(self, data: bytes) -> bytes:
        return data

    def object_type(self) -> Type[Any]:
        return bytes


class AvroDatasetSchemaSerializer(schema.DatasetSchemaVisitor[Dict[str, Any]]):
    """A visitor class that serializes an AvroDatasetSchema as Avro-compatible JSON"""

    def process_schema_field(self, field: schema.SchemaField, avro_type: Union[str, Dict[str, Any]]) -> Dict[str, Any]:
        """Common processing across all field types"""
        # The way to declare nullable fields in Avro schemas is to declare Avro Union (list) types.
        # The default type (usually null) should be listed first:
        # https://avro.apache.org/docs/current/spec.html#Unions
        field_type = avro_type if field.required else ["null", avro_type]
        return {
            "name": field.name,
            "type": field_type,
            "_description": field.description,
            **field.custom_field_tags,
        }

    def process_record_field(self, field: schema.RecordField) -> Dict[str, Any]:
        record_type = {
            "type": "record",
            "fields": [nested_field._accept_visitor(self) for nested_field in field.fields],
            "name": field.name,
        }
        return self.process_schema_field(field, record_type)

    def process_int_field(self, field: schema.IntField) -> Dict[str, Any]:
        return {
            **self.process_schema_field(field, "int"),
        }

    def process_long_field(self, field: schema.LongField) -> Dict[str, Any]:
        return {
            **self.process_schema_field(field, "long"),
        }

    def process_string_field(self, field: schema.StringField) -> Dict[str, Any]:
        return {
            **self.process_schema_field(field, "string"),
        }

    def process_bool_field(self, field: schema.BoolField) -> Dict[str, Any]:
        return {
            **self.process_schema_field(field, "boolean"),
        }

    def process_float_field(self, field: schema.FloatField) -> Dict[str, Any]:
        return {
            **self.process_schema_field(field, "float"),
        }

    def process_double_field(self, field: schema.DoubleField) -> Dict[str, Any]:
        return {
            **self.process_schema_field(field, "double"),
        }

    def process_object_field(self, field: schema.ObjectField) -> Dict[str, Any]:
        return {
            **self.process_schema_field(field, "bytes"),
            "_l5ml_metatype": "object",
            "_is_heavy_pointer": field.is_heavy_pointer,
        }

    def process_array_field(self, field: schema.ArrayField) -> Dict[str, Any]:
        array_type = field.element_field._accept_visitor(self)
        array_type["items"] = array_type["type"]
        array_type["type"] = "array"

        field_type = array_type if field.required else ["null", array_type]
        return {
            "name": field.name,
            "type": field_type,
        }



import multiprocessing
import os
import re
import sys
import tempfile
import time
import unittest
from typing import Optional

from wicker.core.filelock import SimpleUnixFileLock


def write_index_to_file(filepath: str, index: int) -> None:
    # If we don't protect this writing operation with a lock, we will get interleaved
    # writing in the file instead of lines that have "foo-[index]"
    with SimpleUnixFileLock(f"{filepath}.lock"):
        for c in f"foo-{index}\n":
            with open(filepath, "a") as f:
                f.write(c)


class TestFileLock(unittest.TestCase):
    def test_simple_acquire_no_race_conditions(self) -> None:
        with tempfile.NamedTemporaryFile() as lockfile:
            placeholder: Optional[str] = None
            with SimpleUnixFileLock(lockfile.name):
                placeholder = "foo"
            self.assertEqual(placeholder, "foo")

    def test_synchronized_concurrent_writes(self) -> None:
        with tempfile.NamedTemporaryFile("w+") as write_file:
            with multiprocessing.Pool(4) as pool:
                pool.starmap(write_index_to_file, [(write_file.name, i) for i in range(1000)])
            for line in write_file.readlines():
                self.assertTrue(re.match(r"foo-[0-9]+", line))

    def test_timeout(self) -> None:
        with tempfile.NamedTemporaryFile() as lockfile:

            def run_acquire_and_sleep() -> None:
                with SimpleUnixFileLock(lockfile.name):
                    time.sleep(2)

            proc = multiprocessing.Process(target=run_acquire_and_sleep)
            proc.start()

            # Give the other process some time to start up
            time.sleep(1)

            with self.assertRaises(TimeoutError):
                with SimpleUnixFileLock(lockfile.name, timeout_seconds=1):
                    assert False, "Lock acquisition should time out"

    def test_process_dies(self) -> None:
        with tempfile.NamedTemporaryFile() as lockfile:

            def run_acquire_and_die() -> None:
                """Simulate a process dying on some exception while the lock is acquired"""
                # Mute itself to prevent too much noise in stdout/stderr
                with open(os.devnull, "w") as devnull:
                    sys.stdout = devnull
                    sys.stderr = devnull
                    lock = SimpleUnixFileLock(lockfile.name)
                    lock.__enter__()
                    raise ValueError("die")

            # When the child process dies, the fd (and the exclusive lock) should be closed automatically
            proc = multiprocessing.Process(target=run_acquire_and_die)
            proc.start()

            # Give the other process some time to start up
            time.sleep(1)

            placeholder: Optional[str] = None
            with SimpleUnixFileLock(lockfile.name, timeout_seconds=1):
                placeholder = "foo"
            self.assertEqual(placeholder, "foo")



from typing import Any, Dict, List, Optional, Tuple

from wicker.core.definitions import ExampleMetadata
from wicker.core.errors import WickerSchemaException
from wicker.schema import schema, validation


def parse_example(example: Dict[str, Any], schema: schema.DatasetSchema) -> validation.AvroRecord:
    """Parses an example according to the provided schema, converting known types such as
    numpy arrays, torch tensors etc into bytes for storage on disk

    The returned dictionary will be in the same shape as defined in the schema, with keys
    corresponding to the schema names, and values validated against the schema's fields.

    :param example: example to parse
    :type example: Dict[str, Any]
    :param schema: schema to parse against
    :type schema: schema.DatasetSchema
    :return: parsed example
    :rtype: Dict[str, Any]
    """
    parse_example_visitor = ParseExampleVisitor(example, schema)
    return parse_example_visitor.parse_example()


def parse_example_metadata(example: Dict[str, Any], schema: schema.DatasetSchema) -> ExampleMetadata:
    """Parses ExampleMetadata from an example according to the provided schema

    The returned dictionary will be in the same shape as defined in the schema, with keys
    corresponding to the schema names, and values validated against the schema's fields.

    Certain field types will be ignored when parsing examples as they are not considered part of the example metadata:

        1. BytesField
        ... More to be implemented (e.g. AvroNumpyField etc)

    :param example: example to parse
    :type example: Dict[str, Any]
    :param schema: schema to parse against
    :type schema: schema.DatasetSchema
    """
    parse_example_metadata_visitor = ParseExampleMetadataVisitor(example, schema)
    return parse_example_metadata_visitor.parse_example()


# A special exception to indicate that a field should be skipped
class _SkipFieldException(Exception):
    pass


class ParseExampleVisitor(schema.DatasetSchemaVisitor[Any]):
    """schema.DatasetSchemaVisitor class that will validate and transform an Example
    in accordance to a provided schema.DatasetSchema

    The original example is not modified in-place, and we incur the cost of copying
    primitives (e.g. bytes, strings)
    """

    def __init__(self, example: Dict[str, Any], schema: schema.DatasetSchema):
        # Pointers to original data (data should be kept immutable)
        self._schema = schema
        self._example = example

        # Pointers to help keep visitor state during tree traversal
        self._current_data: Any = self._example
        self._current_path: Tuple[str, ...] = tuple()

    def parse_example(self) -> Dict[str, Any]:
        """Parses an example into a form that is suitable for storage in an Avro format"""
        # Since the original input example is non-None, the parsed example will be non-None also
        example: Dict[str, Any] = self._schema.schema_record._accept_visitor(self)
        return example

    def process_record_field(self, field: schema.RecordField) -> Optional[validation.AvroRecord]:
        """Visit an schema.RecordField schema field"""
        val = validation.validate_field_type(self._current_data, dict, field.required, self._current_path)
        if val is None:
            return val

        # Add keys to the example for any non-required fields that were left unset in the raw data
        for optional_field in [f for f in field.fields if not f.required]:
            if optional_field.name not in val:
                val[optional_field.name] = None

        # Validate that data matches schema exactly
        schema_key_names = {nested_field.name for nested_field in field.fields}
        if val.keys() != schema_key_names:
            raise WickerSchemaException(
                f"Error at path {'.'.join(self._current_path)}: "
                f"Example missing keys: {list(schema_key_names - val.keys())} "
                f"and has extra keys: {list(val.keys() - schema_key_names)}"
            )

        # Process nested fields by setting up the visitor's state and visiting each node
        res = {}
        processing_path = self._current_path
        processing_example = self._current_data

        for nested_field in field.fields:
            self._current_path = processing_path + (nested_field.name,)
            self._current_data = processing_example[nested_field.name]
            try:
                res[nested_field.name] = nested_field._accept_visitor(self)
            except _SkipFieldException:
                pass
        return res

    def process_int_field(self, field: schema.IntField) -> Optional[int]:
        return validation.validate_field_type(self._current_data, int, field.required, self._current_path)

    def process_long_field(self, field: schema.LongField) -> Optional[int]:
        return validation.validate_field_type(self._current_data, int, field.required, self._current_path)

    def process_string_field(self, field: schema.StringField) -> Optional[str]:
        return validation.validate_field_type(self._current_data, str, field.required, self._current_path)

    def process_bool_field(self, field: schema.BoolField) -> Optional[bool]:
        return validation.validate_field_type(self._current_data, bool, field.required, self._current_path)

    def process_float_field(self, field: schema.FloatField) -> Optional[float]:
        return validation.validate_field_type(self._current_data, float, field.required, self._current_path)

    def process_double_field(self, field: schema.DoubleField) -> Optional[float]:
        return validation.validate_field_type(self._current_data, float, field.required, self._current_path)

    def process_object_field(self, field: schema.ObjectField) -> Optional[bytes]:
        data = validation.validate_field_type(
            self._current_data, field.codec.object_type(), field.required, self._current_path
        )
        if data is None:
            return None
        return field.codec.validate_and_encode_object(data)

    def process_array_field(self, field: schema.ArrayField) -> Optional[List[Any]]:
        val = validation.validate_field_type(self._current_data, list, field.required, self._current_path)
        if val is None:
            return val

        # Process array elements by setting up the visitor's state and visiting each element
        res = []
        processing_path = self._current_path
        processing_example = self._current_data

        # Arrays may contain None values if the element field declares that it is not required
        for element_index, element in enumerate(processing_example):
            self._current_path = processing_path + (f"elem[{element_index}]",)
            self._current_data = element
            # Allow _SkipFieldExceptions to propagate up and skip this array field
            res.append(field.element_field._accept_visitor(self))
        return res


class ParseExampleMetadataVisitor(ParseExampleVisitor):
    """Specialization of ParseExampleVisitor which skips over certain fields that are now parsed as metadata"""

    def process_object_field(self, field: schema.ObjectField) -> Optional[bytes]:
        # Raise a special error to indicate that this field should be skipped
        raise _SkipFieldException()



"""Classes handling the shuffling of data in S3 when committing a dataset.

When committing a dataset, we sort the data by primary_key before materializing in S3
as Parquet files.

1. The ShuffleJob is the unit of work, and it is just an ordered set of Examples that should
be bundled together in one Parquet file

2. The ShuffleJobFactory produces ShuffleJobs, using a DatasetWriter object to retrieve the
written examples for divvying up as ShuffleJobs.

3. ShuffleWorkers receive the ShuffleJobs and perform the act of retrieving the data and
persisting the data into S3 as Parquet files (one for each ShuffleJob)
"""
from __future__ import annotations

import collections
import concurrent.futures
import dataclasses
import os
import pickle
import tempfile
from typing import Any, Dict, Generator, List, Optional, Tuple

import boto3
import pyarrow as pa
import pyarrow.fs as pafs
import pyarrow.parquet as papq

from wicker.core.column_files import ColumnBytesFileWriter
from wicker.core.definitions import DatasetID, DatasetPartition
from wicker.core.storage import S3DataStorage, S3PathFactory
from wicker.core.writer import DatasetWriterBackend
from wicker.schema import serialization

# Maximum working set for each worker
DEFAULT_WORKER_MAX_WORKING_SET_SIZE = 16384


@dataclasses.dataclass
class ShuffleJob:
    """Represents all the shuffling operations that will happen for a given partition (train/eval/test) on a given
    compute shard."""

    dataset_partition: DatasetPartition
    files: List[Tuple[str, int]]


class ShuffleJobFactory:
    def __init__(
        self,
        writer_backend: DatasetWriterBackend,
        worker_max_working_set_size: int = DEFAULT_WORKER_MAX_WORKING_SET_SIZE,
    ):
        self.writer_backend = writer_backend

        # Factory configurations
        self.worker_max_working_set_size = worker_max_working_set_size

    def build_shuffle_jobs(self, dataset_id: DatasetID) -> Generator[ShuffleJob, None, None]:
        # Initialize with first item
        example_keys = self.writer_backend._metadata_db.scan_sorted(dataset_id)
        try:
            initial_key = next(example_keys)
        except StopIteration:
            return
        job = ShuffleJob(
            dataset_partition=DatasetPartition(dataset_id=dataset_id, partition=initial_key.partition),
            files=[(initial_key.row_data_path, initial_key.row_size)],
        )

        # Yield ShuffleJobs as we accumulate ExampleKeys, where each ShuffleJob is upper-bounded in size by
        # self.worker_max_working_set_size and has all ExampleKeys from the same partition
        for example_key in example_keys:
            if (
                example_key.partition == job.dataset_partition.partition
                and len(job.files) < self.worker_max_working_set_size
            ):
                job.files.append((example_key.row_data_path, example_key.row_size))
                continue

            # Yield job and construct new job to keep iterating
            yield job
            job = ShuffleJob(
                dataset_partition=DatasetPartition(dataset_id=dataset_id, partition=example_key.partition),
                files=[(example_key.row_data_path, example_key.row_size)],
            )
        else:
            yield job


_download_thread_session: Optional[boto3.session.Session] = None
_download_thread_client: Optional[S3DataStorage] = None


def _initialize_download_thread():
    global _download_thread_session
    global _download_thread_client
    if _download_thread_client is None:
        _download_thread_session = boto3.session.Session()
        _download_thread_client = S3DataStorage(session=_download_thread_session)


class ShuffleWorker:
    def __init__(
        self,
        target_rowgroup_bytes_size: int = int(256e6),
        max_worker_threads: int = 16,
        max_memory_usage_bytes: int = int(2e9),
        storage: S3DataStorage = S3DataStorage(),
        s3_path_factory: S3PathFactory = S3PathFactory(),
    ):
        self.target_rowgroup_bytes_size = target_rowgroup_bytes_size
        self.max_worker_threads = max_worker_threads
        self.max_memory_usage_bytes = max_memory_usage_bytes
        self.storage = storage
        self.s3_path_factory = s3_path_factory

    def _download_files(self, job: ShuffleJob) -> Generator[Dict[str, Any], None, None]:
        """Downloads the files in a ShuffleJob, yielding a generator of Dict[str, Any]

        Internally, this function maintains a buffer of lookahead downloads to execute downloads
        in parallel over a ThreadPoolExecutor, up to a maximum of `max_memory_usage_bytes` bytes.

        Args:
            job (ShuffleJob): job to download files for

        Yields:
            Generator[Dict[str, Any], None, None]: stream of Dict[str, Any] from downloading the files
                in order
        """
        # TODO(jchia): Add retries here
        def _download_file(filepath: str) -> Dict[str, Any]:
            assert _download_thread_client is not None
            return pickle.loads(_download_thread_client.fetch_obj_s3(filepath))

        with concurrent.futures.ThreadPoolExecutor(
            self.max_worker_threads, initializer=_initialize_download_thread
        ) as executor:
            buffer: Dict[int, Tuple[concurrent.futures.Future[Dict[str, Any]], int]] = {}
            buffer_size = 0
            buffer_index = 0
            for current_index in range(len(job.files)):
                while buffer_index < len(job.files) and buffer_size < self.max_memory_usage_bytes:
                    filepath, file_size = job.files[buffer_index]
                    buffer[buffer_index] = (executor.submit(_download_file, filepath), file_size)
                    buffer_size += file_size
                    buffer_index += 1
                current_future, current_file_size = buffer[current_index]
                yield current_future.result()
                buffer_size -= current_file_size
                del buffer[current_index]

    def _estimate_target_file_rowgroup_size(
        self,
        job: ShuffleJob,
        target_rowgroup_size_bytes: int = int(256e6),
        min_target_rowgroup_size: int = 16,
    ) -> int:
        """Estimates the number of rows to include in each rowgroup using a target size for the rowgroup

        :param job: job to estimate
        :param target_rowgroup_size_bytes: target size in bytes of a rowgroup, defaults to 256MB
        :param min_target_rowgroup_size: minimum number of rows in a rowgroup
        :return: target number of rows in a rowgroup
        """
        average_filesize = sum([size for _, size in job.files]) / len(job.files)
        return max(min_target_rowgroup_size, int(target_rowgroup_size_bytes / average_filesize))

    def process_job(self, job: ShuffleJob) -> pa.Table:
        # Load dataset schema
        dataset_schema = serialization.loads(
            self.storage.fetch_obj_s3(
                self.s3_path_factory.get_dataset_schema_path(job.dataset_partition.dataset_id)
            ).decode("utf-8")
        )

        # Estimate how many rows to add to each ColumnBytesFile
        target_file_rowgroup_size = self._estimate_target_file_rowgroup_size(job)

        # Initialize data containers to dump into parquet
        heavy_pointer_columns = dataset_schema.get_pointer_columns()
        metadata_columns = dataset_schema.get_non_pointer_columns()
        parquet_metadata: Dict[str, List[Any]] = collections.defaultdict(list)

        # Parse each row, uploading heavy_pointer bytes to S3 and storing only pointers
        # in parquet_metadata
        with ColumnBytesFileWriter(
            self.storage,
            self.s3_path_factory,
            target_file_rowgroup_size=target_file_rowgroup_size,
        ) as writer:
            for data in self._download_files(job):
                for col in metadata_columns:
                    parquet_metadata[col].append(data[col])
                for col in heavy_pointer_columns:
                    loc = writer.add(col, data[col])
                    parquet_metadata[col].append(loc.to_bytes())

        # Save parquet_metadata as a PyArrow Table
        assert len({len(parquet_metadata[col]) for col in parquet_metadata}) == 1, "All columns must have same length"
        return pa.Table.from_pydict(parquet_metadata)


def save_index(
    dataset_name: str,
    dataset_version: str,
    final_indices: Dict[str, pa.Table],
    s3_path_factory: Optional[S3PathFactory] = None,
    s3_storage: Optional[S3DataStorage] = None,
) -> None:
    """Saves a generated final_index into persistent storage

    :param dataset_name: Name of the dataset
    :param dataset_version: Version of the dataset
    :param final_index: Dictionary of pandas dataframes which is the finalized index
    :param pyarrow_filesystem: PyArrow filesystem to use, defaults to None
    :param s3_path_factory: S3PathFactory to use
    :param s3_storage: S3DataStorage to use
    """
    s3_storage = s3_storage if s3_storage is not None else S3DataStorage()
    s3_path_factory = s3_path_factory if s3_path_factory is not None else S3PathFactory()
    for partition_name in final_indices:
        dataset_partition = DatasetPartition(
            dataset_id=DatasetID(
                name=dataset_name,
                version=dataset_version,
            ),
            partition=partition_name,
        )

        parquet_folder = s3_path_factory.get_dataset_partition_path(dataset_partition, s3_prefix=True)
        parquet_path = os.path.join(parquet_folder, "part-0.parquet")

        # Write the Parquet file as one file locally, then upload to S3
        with tempfile.NamedTemporaryFile() as tmpfile:
            pa_table = final_indices[partition_name]
            papq.write_table(
                pa_table,
                tmpfile.name,
                compression="zstd",
                row_group_size=None,
                filesystem=pafs.LocalFileSystem(),
                # We skip writing statistics since it bloats the file, and we don't actually run any queries
                # on the Parquet files that could make use of predicate push-down
                write_statistics=False,
            )
            s3_storage.put_file_s3(
                tmpfile.name,
                parquet_path,
            )

    return None



"""
This file is for S3 storage routines used by S3 datasets

Boto3 S3 documentation links:
- https://boto3.amazonaws.com/v1/documentation/api/latest/guide/index.html
- https://boto3.amazonaws.com/v1/documentation/api/latest/guide/resources.html
- https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html
"""
import io
import logging
import os
import uuid
from typing import Any, Dict, Optional, Tuple
from urllib.parse import urlparse

import boto3
import boto3.session
from botocore.exceptions import ClientError  # type: ignore

from wicker.core.config import get_config
from wicker.core.definitions import DatasetID, DatasetPartition
from wicker.core.filelock import SimpleUnixFileLock

logger = logging.getLogger(__name__)


class S3DataStorage:
    """Storage routines for reading and writing objects in S3"""

    def __init__(self, session: Optional[boto3.session.Session] = None) -> None:
        """Constructor for an S3Storage instance

        Part of the reason to structure S3 routines as a class rather utility functions is to enable
        unit tests to easily mock / patch the S3 client or to utilize the S3 Stubber. Unit tests
        might also find it convenient to mock or patch member functions on instances of this class.
        """
        self.session = boto3.session.Session() if session is None else session
        self.client = self.session.client("s3")

    def __getstate__(self) -> Dict[Any, Any]:
        return {}

    def __setstate__(self, state: Dict[Any, Any]) -> None:
        self.session = boto3.session.Session()
        self.client = self.session.client("s3")
        return

    @staticmethod
    def bucket_key_from_s3_path(s3_path: str) -> Tuple[str, str]:
        """Parses a fully-qualified S3 path such as "s3://hello/world" into the bucket and key

        :param s3_path: fully-qualified S3 path
        :type s3_path: str
        :return: tuple consisting of the S3 bucket and key
        :rtype: Tuple[str, str]
        """
        url_parsed = urlparse(s3_path)
        return url_parsed.netloc, url_parsed.path[1:]

    def check_exists_s3(self, input_path: str) -> bool:
        """Checks if a file exists on S3 under given path

        :param input_path: input file path in S3
        :type input_path: str
        :return: whether or not the file exists in S3
        :rtype: bool
        """
        bucket, key = self.bucket_key_from_s3_path(input_path)
        try:
            self.client.head_object(Bucket=bucket, Key=key)
            return True
        except ClientError:
            return False

    def fetch_file_s3(self, input_path: str, local_prefix: str, timeout_seconds: int = 120) -> str:
        """Fetches a file from S3 to the local machine and skips it if it already exists. This function
        is safe to call concurrently from multiple processes and utilizes a local filelock to block
        parallel downloads such that only one process will perform the download.

        This function assumes the input_path is a valid file in S3.

        :param input_path: input file path in S3
        :param local_prefix: local path that specifies where to download the file
        :param timeout_seconds: number of seconds till timing out on waiting for the file to be downloaded
        :return: local path to the downloaded file
        """
        bucket, key = self.bucket_key_from_s3_path(input_path)
        local_path = os.path.join(local_prefix, key)

        lock_path = local_path + ".lock"
        success_marker = local_path + ".success"
        os.makedirs(os.path.dirname(local_path), exist_ok=True)

        while not os.path.isfile(success_marker):
            with SimpleUnixFileLock(lock_path, timeout_seconds=timeout_seconds):
                if not os.path.isfile(success_marker):

                    # For now, we will only download the file if it has not already been downloaded already.
                    # Long term, we would also add a size check or md5sum comparison against the object in S3.
                    filedir = os.path.split(local_path)[0]
                    os.makedirs(filedir, exist_ok=True)
                    self.client.download_file(bucket, key, local_path)

                    with open(success_marker, "w"):
                        pass

        return local_path

    def fetch_obj_s3(self, input_path: str) -> bytes:
        """Fetches an object from S3 as bytes in memory

        :param input_path: path to object in s3
        :return: bytes of data in file
        """
        bucket, key = self.bucket_key_from_s3_path(input_path)
        bio = io.BytesIO()
        self.client.download_fileobj(bucket, key, bio)
        return bio.getvalue()

    def put_object_s3(self, object_bytes: bytes, s3_path: str) -> None:
        """Upload an object to S3

        :param object_bytes: the object to upload to S3
        :type object_bytes: bytes
        :param s3_path: path to the file in S3
        :type s3_path: str
        """
        # Long term, we would add an md5sum check and short-circuit the upload if they are the same
        bucket, key = self.bucket_key_from_s3_path(s3_path)
        self.client.put_object(Body=object_bytes, Bucket=bucket, Key=key)

    def put_file_s3(self, local_path: str, s3_path: str) -> None:
        """Upload a file to S3

        :param local_path: local path to the file
        :param s3_path: s3 path to dump file to
        """
        bucket, key = self.bucket_key_from_s3_path(s3_path)
        self.client.upload_file(local_path, bucket, key)

    def __eq__(self, other: Any) -> bool:
        # We don't want to use isinstance here to make sure we have the same implementation.
        return super().__eq__(other) and type(self) == type(other)


class S3PathFactory:
    """Factory class for S3 paths

    Our bucket should look like::

        s3://<root_path>
        /__COLUMN_CONCATENATED_FILES__
            - file1
            - file2
        /dataset_name_1
        /dataset_name_2
            /0.0.1
            /0.0.2
            - avro_schema.json
            / assets
                - files added by users
            / partition_1.parquet
            / partition_2.parquet
                - _l5ml_dataset_partition_metadata.json
                - _SUCCESS
                - part-0-attempt-1234.parquet
                - part-1-attempt-2345.parquet
    """

    def __init__(self, s3_root_path: str = None) -> None:
        if s3_root_path is None:
            s3_root_path = get_config().aws_s3_config.s3_datasets_path
        self.root_path = s3_root_path

    def __eq__(self, other: Any) -> bool:
        return super().__eq__(other) and type(self) == type(other) and self.root_path == other.root_path

    def get_dataset_assets_path(self, dataset_id: DatasetID, s3_prefix: bool = True) -> str:
        full_path = os.path.join(
            self.root_path,
            dataset_id.name,
            dataset_id.version,
            "assets",
        )
        if not s3_prefix:
            return full_path.replace("s3://", "")
        return full_path

    def get_dataset_partition_metadata_path(self, data_partition: DatasetPartition, s3_prefix: bool = True) -> str:
        full_path = os.path.join(
            self.get_dataset_partition_path(data_partition),
            "_l5ml_dataset_partition_metadata.json",
        )
        if not s3_prefix:
            return full_path.replace("s3://", "")
        return full_path

    def get_dataset_partition_path(self, data_partition: DatasetPartition, s3_prefix: bool = True) -> str:
        full_path = os.path.join(
            self.root_path,
            data_partition.dataset_id.name,
            data_partition.dataset_id.version,
            f"{data_partition.partition}.parquet",
        )
        if not s3_prefix:
            return full_path.replace("s3://", "")
        return full_path

    def get_dataset_schema_path(self, dataset_id: DatasetID, s3_prefix: bool = True) -> str:
        full_path = os.path.join(
            self.root_path,
            dataset_id.name,
            dataset_id.version,
            "avro_schema.json",
        )
        if not s3_prefix:
            return full_path.replace("s3://", "")
        return full_path

    def get_column_concatenated_bytes_files_path(self, s3_prefix: bool = True) -> str:
        """Gets the path to the root of all column_concatenated_bytes files

        :param s3_prefix: whether to return the s3:// prefix, defaults to True
        :return: path to the column_concatenated_bytes file with the file_id
        """
        full_path = os.path.join(self.root_path, "__COLUMN_CONCATENATED_FILES__")
        if not s3_prefix:
            return full_path.replace("s3://", "")
        return full_path

    def get_column_concatenated_bytes_s3path_from_uuid(self, file_uuid: bytes) -> str:
        columns_root_path = self.get_column_concatenated_bytes_files_path()
        return os.path.join(columns_root_path, str(uuid.UUID(bytes=file_uuid)))

    def get_temporary_row_files_path(self, dataset_id: DatasetID) -> str:
        full_path = os.path.join(
            self.root_path,
            "__TEMP_FILES__",
            dataset_id.name,
            dataset_id.version,
        )
        return full_path



import abc
import os
from typing import Any, Dict, List, Optional

import pyarrow  # type: ignore
import pyarrow.fs as pafs  # type: ignore
import pyarrow.parquet as papq  # type: ignore

from wicker.core.column_files import ColumnBytesFileCache
from wicker.core.config import get_config  # type: ignore
from wicker.core.definitions import DatasetDefinition, DatasetID, DatasetPartition
from wicker.core.storage import S3DataStorage, S3PathFactory
from wicker.schema import dataloading, serialization
from wicker.schema.schema import DatasetSchema

# How long to wait before timing out on filelocks in seconds
FILE_LOCK_TIMEOUT_SECONDS = 240


class AbstractDataset(abc.ABC):
    """Interface for a Map-style (non-streaming) Dataset"""

    @abc.abstractmethod
    def __getitem__(self, idx: int) -> Dict[str, Any]:
        """Gets the ith Example from the given dataset/version/partition"""
        pass

    @abc.abstractmethod
    def __len__(self) -> int:
        """Returns the length of the dataset/version/partition"""
        pass

    @abc.abstractmethod
    def schema(self) -> DatasetSchema:
        """Return the schema of the dataset."""
        pass

    @abc.abstractmethod
    def arrow_table(self) -> pyarrow.Table:
        """Return the pyarrow table with all the metadata fields and pointers of the dataset."""
        pass


class S3Dataset(AbstractDataset):
    """Implementation for a Map-based dataset"""

    def __init__(
        self,
        dataset_name: str,
        dataset_version: str,
        dataset_partition_name: str,
        local_cache_path_prefix: str = os.getenv("TMPDIR", "/tmp"),
        columns_to_load: Optional[List[str]] = None,
        storage: Optional[S3DataStorage] = None,
        s3_path_factory: Optional[S3PathFactory] = None,
        pa_filesystem: Optional[pafs.FileSystem] = None,
        filelock_timeout_seconds: int = FILE_LOCK_TIMEOUT_SECONDS,
        treat_objects_as_bytes: bool = False,
    ):
        """Initializes an S3Dataset

        :param dataset_name: name of the dataset
        :param dataset_version: version of the dataset
        :param dataset_partition_name: partition name
        :param columns_to_load: list of columns to load, defaults to None which loads all columns
        :param data_service: S3DataService instance to use, defaults to None which uses default initializations
        :param filelock_timeout_seconds: number of seconds after which to timeout on waiting for downloads,
            defaults to FILE_LOCK_TIMEOUT_SECONDS
        :param treat_objects_as_bytes: If set, don't try to decode ObjectFields and keep them as binary data.
        """
        super().__init__()
        self._columns_to_load: Optional[List[str]] = columns_to_load
        self._treat_objects_as_bytes = treat_objects_as_bytes
        self._schema: Optional[DatasetSchema] = None
        self._arrow_table: Optional[pyarrow.Table] = None

        self._local_cache_path_prefix = local_cache_path_prefix
        self._filelock_timeout_seconds = filelock_timeout_seconds
        self._storage = storage if storage is not None else S3DataStorage()
        self._s3_path_factory = s3_path_factory if s3_path_factory is not None else S3PathFactory()
        self._column_bytes_file_cache = ColumnBytesFileCache(
            local_cache_path_prefix=local_cache_path_prefix,
            filelock_timeout_seconds=filelock_timeout_seconds,
            path_factory=self._s3_path_factory,
            storage=self._storage,
        )
        self._pa_filesystem = (
            pafs.S3FileSystem(region=get_config().aws_s3_config.region) if pa_filesystem is None else pa_filesystem
        )

        self._dataset_id = DatasetID(name=dataset_name, version=dataset_version)
        self._partition = DatasetPartition(dataset_id=self._dataset_id, partition=dataset_partition_name)
        self._dataset_definition = DatasetDefinition(
            self._dataset_id,
            schema=self.schema(),
        )

    def schema(self) -> DatasetSchema:
        if self._schema is None:
            schema_path = self._s3_path_factory.get_dataset_schema_path(self._dataset_id)
            local_path = self._storage.fetch_file_s3(
                schema_path, self._local_cache_path_prefix, timeout_seconds=self._filelock_timeout_seconds
            )
            with open(local_path, "rb") as f:
                self._schema = serialization.loads(
                    f.read().decode("utf-8"), treat_objects_as_bytes=self._treat_objects_as_bytes
                )
        return self._schema

    def arrow_table(self) -> pyarrow.Table:
        path = self._s3_path_factory.get_dataset_partition_path(self._partition, s3_prefix=False)
        if not self._arrow_table:
            self._arrow_table = papq.read_table(path, filesystem=self._pa_filesystem)
        return self._arrow_table

    def __len__(self) -> int:
        return len(self.arrow_table())

    def __getitem__(self, idx: int) -> Dict[str, Any]:
        tbl = self.arrow_table()
        columns = self._columns_to_load if self._columns_to_load is not None else tbl.column_names
        row = {col: tbl[col][idx].as_py() for col in columns}
        return dataloading.load_example(
            self._column_bytes_file_cache.resolve_pointers(row, self.schema()),
            self.schema(),
        )



import os
import tempfile
import unittest
import uuid
from unittest.mock import MagicMock

from wicker.core.column_files import ColumnBytesFileLocationV1, ColumnBytesFileWriter
from wicker.core.definitions import DatasetID, DatasetPartition
from wicker.core.storage import S3PathFactory
from wicker.testing.storage import FakeS3DataStorage

FAKE_SHARD_ID = "fake_shard_id"
FAKE_DATA_PARTITION = DatasetPartition(dataset_id=DatasetID(name="name", version="0.0.1"), partition="partition")
FAKE_BYTES = b"foo"
FAKE_BYTES2 = b"0123456789"
FAKE_COL = "col0"
FAKE_COL2 = "col1"


class TestColumnBytesFileWriter(unittest.TestCase):
    def test_write_empty(self) -> None:
        mock_storage = MagicMock()
        path_factory = S3PathFactory()
        with ColumnBytesFileWriter(
            storage=mock_storage,
            s3_path_factory=path_factory,
        ):
            pass
        mock_storage.put_object_s3.assert_not_called()

    def test_write_one_column_one_row(self) -> None:
        path_factory = S3PathFactory()
        mock_storage = MagicMock()
        with ColumnBytesFileWriter(
            storage=mock_storage,
            s3_path_factory=path_factory,
        ) as ccb:
            info = ccb.add(FAKE_COL, FAKE_BYTES)
            self.assertEqual(info.byte_offset, 0)
            self.assertEqual(info.data_size, len(FAKE_BYTES))
        mock_storage.put_file_s3.assert_called_once_with(
            unittest.mock.ANY,
            os.path.join(
                path_factory.get_column_concatenated_bytes_files_path(),
                str(info.file_id),
            ),
        )

    def test_write_one_column_multi_row(self) -> None:
        path_factory = S3PathFactory()
        mock_storage = MagicMock()
        with ColumnBytesFileWriter(
            storage=mock_storage,
            s3_path_factory=path_factory,
        ) as ccb:
            info = ccb.add(FAKE_COL, FAKE_BYTES)
            self.assertEqual(info.byte_offset, 0)
            self.assertEqual(info.data_size, len(FAKE_BYTES))

            next_info = ccb.add(FAKE_COL, FAKE_BYTES)
            self.assertEqual(next_info.byte_offset, len(FAKE_BYTES))
            self.assertEqual(next_info.data_size, len(FAKE_BYTES))
        mock_storage.put_file_s3.assert_called_once_with(
            unittest.mock.ANY,
            os.path.join(
                path_factory.get_column_concatenated_bytes_files_path(),
                str(info.file_id),
            ),
        )

    def test_write_multi_column_multi_row(self) -> None:
        path_factory = S3PathFactory()
        mock_storage = MagicMock()
        with ColumnBytesFileWriter(
            storage=mock_storage,
            s3_path_factory=path_factory,
        ) as ccb:
            info1 = ccb.add(FAKE_COL, FAKE_BYTES)
            self.assertEqual(info1.byte_offset, 0)
            self.assertEqual(info1.data_size, len(FAKE_BYTES))

            info1 = ccb.add(FAKE_COL, FAKE_BYTES)
            self.assertEqual(info1.byte_offset, len(FAKE_BYTES))
            self.assertEqual(info1.data_size, len(FAKE_BYTES))

            info2 = ccb.add(FAKE_COL2, FAKE_BYTES)
            self.assertEqual(info2.byte_offset, 0)
            self.assertEqual(info2.data_size, len(FAKE_BYTES))

            info2 = ccb.add(FAKE_COL2, FAKE_BYTES)
            self.assertEqual(info2.byte_offset, len(FAKE_BYTES))
            self.assertEqual(info2.data_size, len(FAKE_BYTES))

            info2 = ccb.add(FAKE_COL2, FAKE_BYTES)
            self.assertEqual(info2.byte_offset, len(FAKE_BYTES) * 2)
            self.assertEqual(info2.data_size, len(FAKE_BYTES))

        mock_storage.put_file_s3.assert_any_call(
            unittest.mock.ANY,
            os.path.join(
                path_factory.get_column_concatenated_bytes_files_path(),
                str(info1.file_id),
            ),
        )
        mock_storage.put_file_s3.assert_any_call(
            unittest.mock.ANY,
            os.path.join(
                path_factory.get_column_concatenated_bytes_files_path(),
                str(info2.file_id),
            ),
        )

    def test_write_large_file(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            path_factory = S3PathFactory()
            storage = FakeS3DataStorage(tmpdir=tmpdir)
            with ColumnBytesFileWriter(storage=storage, s3_path_factory=path_factory, target_file_size=10) as ccb:
                info1 = ccb.add(FAKE_COL, FAKE_BYTES)
                self.assertEqual(info1.byte_offset, 0)
                self.assertEqual(info1.data_size, len(FAKE_BYTES))
                info1 = ccb.add(FAKE_COL, FAKE_BYTES)
                self.assertEqual(info1.byte_offset, len(FAKE_BYTES))
                self.assertEqual(info1.data_size, len(FAKE_BYTES))
                info1 = ccb.add(FAKE_COL, FAKE_BYTES)
                self.assertEqual(info1.byte_offset, len(FAKE_BYTES) * 2)
                self.assertEqual(info1.data_size, len(FAKE_BYTES))
                info1 = ccb.add(FAKE_COL, FAKE_BYTES)
                self.assertEqual(info1.byte_offset, len(FAKE_BYTES) * 3)
                self.assertEqual(info1.data_size, len(FAKE_BYTES))
                info2 = ccb.add(FAKE_COL, FAKE_BYTES)
                self.assertEqual(info2.byte_offset, 0)
                self.assertEqual(info2.data_size, len(FAKE_BYTES))
                self.assertNotEqual(info1.file_id, info2.file_id)

                info1 = ccb.add(FAKE_COL2, FAKE_BYTES2)
                self.assertEqual(info1.byte_offset, 0)
                self.assertEqual(info1.data_size, len(FAKE_BYTES2))
                info2 = ccb.add(FAKE_COL2, FAKE_BYTES2)
                self.assertEqual(info2.byte_offset, 0)
                self.assertEqual(info2.data_size, len(FAKE_BYTES2))
                self.assertNotEqual(info1.file_id, info2.file_id)
                info3 = ccb.add(FAKE_COL2, FAKE_BYTES2)
                self.assertEqual(info3.byte_offset, 0)
                self.assertEqual(info3.data_size, len(FAKE_BYTES2))
                self.assertNotEqual(info2.file_id, info3.file_id)

    def test_write_manyrows_file(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            path_factory = S3PathFactory()
            storage = FakeS3DataStorage(tmpdir=tmpdir)
            with ColumnBytesFileWriter(
                storage=storage, s3_path_factory=path_factory, target_file_rowgroup_size=1
            ) as ccb:
                info1 = ccb.add(FAKE_COL, FAKE_BYTES)
                self.assertEqual(info1.byte_offset, 0)
                self.assertEqual(info1.data_size, len(FAKE_BYTES))
                info2 = ccb.add(FAKE_COL, FAKE_BYTES)
                self.assertEqual(info2.byte_offset, 0)
                self.assertEqual(info2.data_size, len(FAKE_BYTES))
                self.assertNotEqual(info1.file_id, info2.file_id)

                info1 = ccb.add(FAKE_COL2, FAKE_BYTES2)
                self.assertEqual(info1.byte_offset, 0)
                self.assertEqual(info1.data_size, len(FAKE_BYTES2))
                info2 = ccb.add(FAKE_COL2, FAKE_BYTES2)
                self.assertEqual(info2.byte_offset, 0)
                self.assertEqual(info2.data_size, len(FAKE_BYTES2))
                self.assertNotEqual(info1.file_id, info2.file_id)
                info3 = ccb.add(FAKE_COL2, FAKE_BYTES2)
                self.assertEqual(info3.byte_offset, 0)
                self.assertEqual(info3.data_size, len(FAKE_BYTES2))
                self.assertNotEqual(info2.file_id, info3.file_id)


class TestCCBInfo(unittest.TestCase):
    def test_to_string(self) -> None:
        ccb_info = ColumnBytesFileLocationV1(uuid.uuid4(), 100, 100)
        ccb_info_parsed = ColumnBytesFileLocationV1.from_bytes(ccb_info.to_bytes())
        self.assertEqual(ccb_info, ccb_info_parsed)



"""Spark plugin for writing a dataset with Spark only (no external metadata database required)

This plugin does an expensive global sorting step using Spark, which could be prohibitive
for large datasets.
"""
from __future__ import annotations

from typing import Any, Dict, Iterable, Optional, Tuple

import pyarrow as pa
import pyarrow.compute as pc

try:
    import pyspark
except ImportError:
    raise RuntimeError(
        "pyspark is not detected in current environment, install Wicker with extra arguments:"
        " `pip install wicker[spark]`"
    )

from operator import add

from wicker import schema as schema_module
from wicker.core.definitions import DatasetID
from wicker.core.errors import WickerDatastoreException
from wicker.core.persistance import AbstractDataPersistor
from wicker.core.storage import S3DataStorage, S3PathFactory
from wicker.schema import serialization

SPARK_PARTITION_SIZE = 256
MAX_COL_FILE_NUMROW = 50  # TODO(isaak-willett): Magic number, we should derive this based on row size

PrimaryKeyTuple = Tuple[Any, ...]
UnparsedExample = Dict[str, Any]
ParsedExample = Dict[str, Any]
PointerParsedExample = Dict[str, Any]


def persist_wicker_dataset(
    dataset_name: str,
    dataset_version: str,
    dataset_schema: schema_module.DatasetSchema,
    rdd: pyspark.rdd.RDD[Tuple[str, UnparsedExample]],
    s3_storage: S3DataStorage = S3DataStorage(),
    s3_path_factory: S3PathFactory = S3PathFactory(),
) -> Optional[Dict[str, int]]:
    """
    Persist wicker dataset public facing api function, for api consistency.
    :param dataset_name: name of dataset persisted
    :type dataset_name: str
    :param dataset_version: version of dataset persisted
    :type dataset_version: str
    :param dataset_schema: schema of dataset to be persisted
    :type dataset_schema: DatasetSchema
    :param rdd: rdd of data to persist
    :type rdd: RDD
    :param s3_storage: s3 storage abstraction
    :type s3_storage: S3DataStorage
    :param s3_path_factory: s3 path abstraction
    :type s3_path_factory: S3PathFactory
    """
    return SparkPersistor(s3_storage, s3_path_factory).persist_wicker_dataset(
        dataset_name, dataset_version, dataset_schema, rdd
    )


class SparkPersistor(AbstractDataPersistor):
    def __init__(
        self,
        s3_storage: S3DataStorage = S3DataStorage(),
        s3_path_factory: S3PathFactory = S3PathFactory(),
    ) -> None:
        """
        Init a SparkPersistor

        :param s3_storage: The storage abstraction for S3
        :type s3_storage: S3DataStore
        :param s3_path_factory: The path factory for generating s3 paths
                                based on dataset name and version
        :type s3_path_factory: S3PathFactory
        """
        super().__init__(s3_storage, s3_path_factory)

    def persist_wicker_dataset(
        self,
        dataset_name: str,
        dataset_version: str,
        schema: schema_module.DatasetSchema,
        rdd: pyspark.rdd.RDD[Tuple[str, UnparsedExample]],
    ) -> Optional[Dict[str, int]]:
        """
        Persist the current rdd dataset defined by name, version, schema, and data.
        """
        # check if variables have been set ie: not None
        if (
            not isinstance(dataset_name, str)
            or not isinstance(dataset_version, str)
            or not isinstance(schema, schema_module.DatasetSchema)
            or not isinstance(rdd, pyspark.rdd.RDD)
        ):
            raise ValueError("Current dataset variables not all set, set all to proper not None values")

        # define locally for passing to spark rdd ops, breaks if relying on self
        # since it passes to spark engine and we lose self context
        s3_storage = self.s3_storage
        s3_path_factory = self.s3_path_factory
        parse_row = self.parse_row
        get_row_keys = self.get_row_keys
        persist_wicker_partition = self.persist_wicker_partition
        save_partition_tbl = self.save_partition_tbl

        # put the schema up on to s3
        schema_path = s3_path_factory.get_dataset_schema_path(DatasetID(name=dataset_name, version=dataset_version))
        s3_storage.put_object_s3(serialization.dumps(schema).encode("utf-8"), schema_path)

        # parse the rows and ensure validation passes, ie: rows actual data matches expected types
        rdd0 = rdd.mapValues(lambda row: parse_row(row, schema))  # type: ignore

        # Make sure to cache the RDD to ease future computations, since it seems that sortBy and zipWithIndex
        # trigger actions and we want to avoid recomputing the source RDD at all costs
        rdd0 = rdd0.cache()
        dataset_size = rdd0.count()

        rdd1 = rdd0.keyBy(lambda row: get_row_keys(row, schema))

        # Sort RDD by keys
        rdd2: pyspark.rdd.RDD[Tuple[Tuple[Any, ...], Tuple[str, ParsedExample]]] = rdd1.sortByKey(
            # TODO(jchia): Magic number, we should derive this based on row size
            numPartitions=max(1, dataset_size // SPARK_PARTITION_SIZE),
            ascending=True,
        )

        def set_partition(iterator: Iterable[PrimaryKeyTuple]) -> Iterable[int]:
            key_set = set(iterator)
            yield len(key_set)

        # the number of unique keys in rdd partitions
        # this is softer check than collecting all the keys in all partitions to check uniqueness
        rdd_key_count: int = rdd2.map(lambda x: x[0]).mapPartitions(set_partition).reduce(add)
        num_unique_keys = rdd_key_count
        if dataset_size != num_unique_keys:
            raise WickerDatastoreException(
                f"""Error: dataset examples do not have unique primary key tuples.
                Dataset has has {dataset_size} examples but {num_unique_keys} unique primary keys"""
            )

        # persist the spark partition to S3Storage
        rdd3 = rdd2.values()

        rdd4 = rdd3.mapPartitions(
            lambda spark_iterator: persist_wicker_partition(
                spark_iterator,
                schema,
                s3_storage,
                s3_path_factory,
                target_max_column_file_numrows=MAX_COL_FILE_NUMROW,
            )
        )

        # combine the rdd by the keys in the pyarrow table
        rdd5 = rdd4.combineByKey(
            createCombiner=lambda data: pa.Table.from_pydict(
                {col: [data[col]] for col in schema.get_all_column_names()}
            ),
            mergeValue=lambda tbl, data: pa.Table.from_batches(
                [
                    *tbl.to_batches(),  # type: ignore
                    *pa.Table.from_pydict({col: [data[col]] for col in schema.get_all_column_names()}).to_batches(),
                ]
            ),
            mergeCombiners=lambda tbl1, tbl2: pa.Table.from_batches(
                [
                    *tbl1.to_batches(),  # type: ignore
                    *tbl2.to_batches(),  # type: ignore
                ]
            ),
        )

        # create the partition tables
        rdd6 = rdd5.mapValues(
            lambda pa_tbl: pc.take(
                pa_tbl,
                pc.sort_indices(
                    pa_tbl,
                    sort_keys=[(pk, "ascending") for pk in schema.primary_keys],
                ),
            )
        )

        # save the parition table to s3
        rdd7 = rdd6.map(
            lambda partition_table: save_partition_tbl(
                partition_table, dataset_name, dataset_version, s3_storage, s3_path_factory
            )
        )
        written = rdd7.collect()

        return {partition: size for partition, size in written}

    @staticmethod
    def get_row_keys(
        partition_data_tup: Tuple[str, ParsedExample], schema: schema_module.DatasetSchema
    ) -> PrimaryKeyTuple:
        """
        Get the keys of a row based on the parition tuple and the data schema.

        :param partition_data_tup: Tuple of partition id and ParsedExample row
        :type partition_data_tup: Tuple[str, ParsedExample]
        :return: Tuple of primary key values from parsed row and schema
        :rtype: PrimaryKeyTuple
        """
        partition, data = partition_data_tup
        return (partition,) + tuple(data[pk] for pk in schema.primary_keys)



import json
from typing import Any, Dict, List, Type

from wicker.schema import codecs


class Vector:
    def __init__(self, data: List[int]):
        self.data = data

    def __eq__(self, other: Any) -> bool:
        return super().__eq__(other) and self.data == other.data


class VectorCodec(codecs.Codec):
    def __init__(self, compression_method: int) -> None:
        self.compression_method = compression_method

    @staticmethod
    def _codec_name() -> str:
        return "VectorCodec"

    def save_codec_to_dict(self) -> Dict[str, Any]:
        return {"compression_method": self.compression_method}

    @staticmethod
    def load_codec_from_dict(data: Dict[str, Any]) -> codecs.Codec:
        return VectorCodec(compression_method=data["compression_method"])

    def validate_and_encode_object(self, obj: Vector) -> bytes:
        # Inefficient but simple encoding method for testing.
        return json.dumps(obj.data).encode("utf-8")

    def decode_object(self, data: bytes) -> Vector:
        return Vector(json.loads(data.decode()))

    def object_type(self) -> Type[Any]:
        return Vector



import copy
import os
import random
import tempfile
import unittest
import uuid

import pyarrow.parquet as papq
from pyspark.sql import SparkSession

from wicker import schema
from wicker.core.config import get_config
from wicker.core.errors import WickerDatastoreException
from wicker.plugins.spark import persist_wicker_dataset
from wicker.testing.storage import FakeS3DataStorage

DATASET_NAME = "dataset"
DATASET_VERSION = "0.0.1"
TEST_ROWS_NUM = 10000
SCHEMA = schema.DatasetSchema(
    primary_keys=["bar", "foo"],
    fields=[
        schema.IntField("foo"),
        schema.StringField("bar"),
        schema.BytesField("bytescol"),
    ],
)
EXAMPLES = [
    (
        "train" if i % 2 == 0 else "test",
        {
            "foo": random.randint(0, TEST_ROWS_NUM),
            "bar": str(uuid.uuid4()),
            "bytescol": b"0",
        },
    )
    for i in range(TEST_ROWS_NUM)
]
# Examples with a duplicated key
EXAMPLES_DUPES = copy.deepcopy(EXAMPLES)
EXAMPLES_DUPES[5000] = EXAMPLES_DUPES[0]


class LocalWritingTestCase(unittest.TestCase):
    def assert_written_correctness(self, tmpdir: str, row_num: int = TEST_ROWS_NUM) -> None:
        """Asserts that all files are written as expected by the L5MLDatastore"""
        # Check that files are correctly written locally by Spark/Parquet with a _SUCCESS marker file
        prefix = get_config().aws_s3_config.s3_datasets_path.replace("s3://", "")
        self.assertIn(DATASET_NAME, os.listdir(os.path.join(tmpdir, prefix)))
        self.assertIn(DATASET_VERSION, os.listdir(os.path.join(tmpdir, prefix, DATASET_NAME)))
        for partition in ["train", "test"]:
            columns_path = os.path.join(tmpdir, prefix, "__COLUMN_CONCATENATED_FILES__")
            all_read_bytes = b""
            for filename in os.listdir(columns_path):
                concatenated_bytes_filepath = os.path.join(columns_path, filename)
                with open(concatenated_bytes_filepath, "rb") as bytescol_file:
                    all_read_bytes += bytescol_file.read()
            self.assertEqual(all_read_bytes, b"0" * row_num)

            # Load parquet file and assert ordering of primary_key
            self.assertIn(
                f"{partition}.parquet", os.listdir(os.path.join(tmpdir, prefix, DATASET_NAME, DATASET_VERSION))
            )
            tbl = papq.read_table(os.path.join(tmpdir, prefix, DATASET_NAME, DATASET_VERSION, f"{partition}.parquet"))
            foobar = [(barval.as_py(), fooval.as_py()) for fooval, barval in zip(tbl["foo"], tbl["bar"])]
            self.assertEqual(foobar, sorted(foobar))

    def test_simple_schema_local_writing(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            fake_storage = FakeS3DataStorage(tmpdir=tmpdir)
            spark_session = SparkSession.builder.appName("test").master("local[*]")
            spark = spark_session.getOrCreate()
            sc = spark.sparkContext
            rdd = sc.parallelize(copy.deepcopy(EXAMPLES), 100)
            persist_wicker_dataset(DATASET_NAME, DATASET_VERSION, SCHEMA, rdd, fake_storage)
            self.assert_written_correctness(tmpdir)

    def test_dupe_primary_keys_raises_exception(self) -> None:
        with self.assertRaises(WickerDatastoreException) as e:
            with tempfile.TemporaryDirectory() as tmpdir:
                fake_storage = FakeS3DataStorage(tmpdir=tmpdir)
                spark_session = SparkSession.builder.appName("test").master("local[*]")
                spark = spark_session.getOrCreate()
                sc = spark.sparkContext
                rdd = sc.parallelize(copy.deepcopy(EXAMPLES_DUPES), 100)
                persist_wicker_dataset(DATASET_NAME, DATASET_VERSION, SCHEMA, rdd, fake_storage)

            self.assertIn(
                "Error: dataset examples do not have unique primary key tuples",
                str(e.exception),
            )

    def test_simple_schema_local_writing_4_row_dataset(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            small_row_cnt = 4
            fake_storage = FakeS3DataStorage(tmpdir=tmpdir)
            spark_session = SparkSession.builder.appName("test").master("local[*]")
            spark = spark_session.getOrCreate()
            sc = spark.sparkContext
            rdd = sc.parallelize(copy.deepcopy(EXAMPLES)[:small_row_cnt], 1)
            persist_wicker_dataset(DATASET_NAME, DATASET_VERSION, SCHEMA, rdd, fake_storage)
            self.assert_written_correctness(tmpdir, small_row_cnt)
